# coding: utf-8

"""

    Tools to Unify Image Data Against Other BioVida APIs
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

"""
# Note: To use these tools, see ``biovida.unify_domains.unify_against_images()``
import numpy as np
import pandas as pd
from copy import deepcopy
from collections import defaultdict

# Import Interfaces
from biovida.genomics.disgenet_interface import DisgenetInterface
from biovida.diagnostics.disease_ont_interface import DiseaseOntInterface
from biovida.diagnostics.disease_symptoms_interface import DiseaseSymptomsInterface

# Support Tools
from biovida.support_tools.support_tools import tqdm, is_int, items_null

# Image Tools
from biovida.images._image_tools import try_fuzzywuzzy_import

# Open-i Specific Tools
from biovida.images._interface_support.openi.openi_support_tools import possible_openi_image_processing_cols


# ----------------------------------------------------------------------------------------------------------
# Interface Integration
# ----------------------------------------------------------------------------------------------------------


class _ImagesInterfaceIntegration(object):
    """

    Tools to Unify BioVida Image Interfaces

    """

    def __init__(self):
        self._additional_columns = list()

    def _add_additional_columns(self, db, list_of_columns, skip=None):
        """

        Add ``self._additional_columns`` to ``db`` and ``list_of_columns``

        :param list_of_columns: as present in ``_open_i_prep()`` or ``_cancer_image_prep()``.
        :type list_of_columns: ``list``
        :param db: as passed to ``_open_i_prep()`` or ``_cancer_image_prep()``.
        :type db: ``Pandas DataFrame``
        :param skip: items in ``self._additional_columns`` to refrain from adding to ``db``.
        :type skip: ``list`` or ``None``
        :return: see description.
        :rtype: ``tuple``
        """
        for c in self._additional_columns:
            if skip is None or c not in skip:
                list_of_columns.append(c)
                if c not in db.columns:
                    db[c] = [None] * db.shape[0]
        return db, list_of_columns

    def _open_i_prep(self, db):
        """

        A tool to clean and standardize a database from a ``OpeniInterface`` instance.

        :param db: a database from an ``OpeniInterface`` instance.
        :type db: ``Pandas DataFrame``
        :return: a cleaned and standardize ``db``.
        :rtype: ``Pandas DataFrame``
        """
        # Deep copy the input to prevent mutating the original in memory.
        db_cln = db.copy(deep=True)

        # Column which provides a guess, based on the text, on which imaging modality created the image.
        db_cln['modality_best_guess'] = db_cln.apply(
            lambda x: x['imaging_modality_from_text'] if isinstance(x['imaging_modality_from_text'], str) else x[
                'modality_full'], axis=1)

        # Define columns to keep
        openi_columns = ['abstract', 'article_type', 'image_id_short', 'image_caption',
                         'modality_best_guess', 'age', 'sex', 'diagnosis', 'query', 'pull_time']

        openi_col_rename = {'diagnosis': 'disease'}

        # Allow for cases where images have not been downloaded.
        if 'cached_images_path' in db_cln.columns:
            openi_columns.append('cached_images_path')
            db_cln['cached_images_path'] = db_cln['cached_images_path'].map(
                lambda x: tuple([x]) if not isinstance(x, tuple) else x, na_action='ignore')

        if len(self._additional_columns):
            db_cln, openi_columns = self._add_additional_columns(db_cln, openi_columns)

        db_cln['article_type'] = db_cln['article_type'].replace({'encounter': 'case_report',
                                                                 'image_id_short': 'image_id'})

        # Define subsection based on `openi_columns`
        openi_subsection = db_cln[openi_columns]

        openi_subsection['source_api'] = ['openi'] * openi_subsection.shape[0]
        return openi_subsection.rename(columns=openi_col_rename)

    def _image_processing_prep(self, db):
        """

        A tool to clean and standardize a database from a ``OpeniImageProcessing`` instance.

        :param db: a database from a ``OpeniImageProcessing`` instance.
        :type db: ``Pandas DataFrame``
        :return: a cleaned and standardize ``db``.
        :rtype: ``Pandas DataFrame``
        """
        # Note: if the ``OpeniImageProcessing`` class is updated to handle
        # instances other than ``OpeniInterface``, this approach will need to be changed.
        return self._open_i_prep(db)

    def _cancer_image_prep(self, db):
        """

        A tool to clean and standardize a database from a ``CancerImageInterface`` instance.

        :param db: a database from a ``CancerImageInterface`` instance.
        :type db: ``Pandas DataFrame``
        :return: a cleaned and standardize ``db``.
        :rtype: ``Pandas DataFrame``
        """
        # Define columns to keep
        cancer_image_columns = ['series_instance_uid', 'series_description', 'modality_full', 'age',
                                'sex', 'article_type', 'cancer_type', 'query', 'pull_time']

        # Column name changes (based on ``_open_i_prep()``).
        cancer_image_col_rename = {'series_instance_uid': 'image_id',
                                   'series_description': 'image_caption',
                                   'cached_dicom_images_path': 'raw',
                                   'modality_full': 'modality_best_guess',
                                   'cancer_type': 'disease'}

        # Allow for cases where images have not been downloaded,
        # i.e., absent 'cached_images_path' and 'cached_dicom_images_path' columns.
        if 'cached_images_path' in db.columns:
            cancer_image_columns.append('cached_images_path')
        if 'cached_dicom_images_path' in db.columns:
            cancer_image_columns.append('cached_dicom_images_path')
            cancer_image_col_rename['cached_dicom_images_path'] = 'source_images_path'
            additional_columns_skip = 'source_images_path'
        else:
            additional_columns_skip = None

        # Deep copy the input to prevent mutating the original in memory.
        db_cln = db.copy(deep=True)

        if len(self._additional_columns):
            db_cln, cancer_image_columns = self._add_additional_columns(db=db_cln,
                                                                        list_of_columns=cancer_image_columns,
                                                                        skip=additional_columns_skip)

        # Define subsection based on `cancer_image_columns`
        cancer_image_subsection = db_cln[cancer_image_columns]

        cancer_image_subsection['abstract'] = np.NaN
        cancer_image_subsection['source_api'] = ['tcia'] * cancer_image_subsection.shape[0]
        return cancer_image_subsection.rename(columns=cancer_image_col_rename)

    @property
    def _prep_class_dict(self):
        """

        Return a dictionary which maps image interface classes to
        the methods designed to handle them.

        :return: a dictionary mapping class names to functions.
        :rtype: ``dict``
        """
        return {'OpeniInterface': self._open_i_prep,
                'CancerImageInterface': self._cancer_image_prep,
                'OpeniImageProcessing': self._image_processing_prep}

    def integration(self, instances, db_to_extract):
        """

        Standardize instances.

        This method yields a single dataframe with the following columns:

         - 'abstract'*
         - 'image_id_short'
         - 'image_caption'
         - 'modality_best_guess'
         - 'age'
         - 'sex'
         - 'disease'
         - 'query'
         - 'pull_time'
         - 'harvest_success'
         - 'files_path'
         - 'source_api'

         *NOTE: this column will be dropped after passing through ``_DiseaseSymptomsIntegration().integration()``.

        :param instances: any one of ``OpeniInterface``, ``CancerImageInterface`` or ``OpeniImageProcessing``, or some
                           combination inside an iterable.
        :type instances: ``list``, ``tuple``, ``OpeniInterface``, ``CancerImageInterface`` or ``OpeniImageProcessing``.
        :param db_to_extract: the database to use. Must be one of: 'records_db', 'cache_records_db'.
        :type db_to_extract: ``str``
        :return: standardize instances
        :rtype: ``Pandas DataFrame``
        """
        instances_types = [type(i).__name__ for i in instances]
        if 'OpeniImageProcessing' in instances_types:
            self._additional_columns += possible_openi_image_processing_cols
        if 'CancerImageInterface' in instances_types:
            self._additional_columns += ['source_images_path']

        frames = list()
        for class_instance in instances:
            interface_name = type(class_instance).__name__
            func = self._prep_class_dict[interface_name]
            if interface_name == 'OpeniImageProcessing':
                database = getattr(class_instance, "image_dataframe")
            else:
                database = getattr(class_instance, db_to_extract)
            if not isinstance(database, pd.DataFrame):
                raise ValueError("The {0} instance's '{1}' database must be of type DataFrame,\nnot "
                                 "'{2}'.".format(interface_name, db_to_extract, type(database).__name__))
            frames.append(func(database))

        self._additional_columns = list()  # reset

        combined_df = pd.concat(frames, ignore_index=True)
        combined_df['disease'] = combined_df['disease'].map(lambda x: x.lower() if isinstance(x, str) else x)

        return combined_df.fillna(np.NaN)


# ----------------------------------------------------------------------------------------------------------
# Disease Ontology Integration
# ----------------------------------------------------------------------------------------------------------


class _DiseaseOntologyIntegration(object):
    """

    Integration of Disease Ontology data.

    :param cache_path: location of the BioVida cache. If one does not exist in this location, one will created.
    Default to ``None`` (which will generate a cache in the home folder).
    :type cache_path: ``str`` or ``None``
    :param verbose: If ``True``, print notice when downloading database. Defaults to ``True``.
    :type verbose: ``bool``
    """

    @staticmethod
    def _dis_ont_dict_gen(ontology_df):
        """

        Convert the information obtained from ``DiseaseOntInterface().pull()`` into:

        - a nested dictionary with ``ontology_df``'s 'name' column as the outer key (``ont_name_dict``).
          Form: ``{'name': {'disease_family' ('is_a'): tuple or None,
                            'disease_synonym': tuple or None,
                            'diagnosis_definition' ('def'): str or None},
                  ...}``

        - the keys of the nested dictionaries in ``ont_name_dict``

        - a dictionary with 'disease_synonym' as keys and related names in a list:
          Form ``{'disease_synonym': ['name', 'name', ...], ...}``

        :param ontology_df: yield of ``DiseaseOntInterface().pull()``
        :type ontology_df: ``Pandas DataFrame``
        :return: see method description.
        :rtype: ``tuple``
        """
        ont_name_dict, ont_disease_synonym_dict = dict(), dict()
        ont_name_dict_nest_keys = ('disease_family', 'disease_synonym', 'disease_definition')

        def str_split(s, split_on='; '):
            return tuple(s.split(split_on)) if isinstance(s, str) else s

        # ToDo: change to iterrows().
        for name, is_a, disease_synonym, defn in zip(*[ontology_df[c] for c in ('name', 'is_a', 'synonym', 'def')]):
            disease_synonym_split = str_split(disease_synonym)
            if not items_null(name):
                # Update `ont_name_dict`
                ont_name_dict[name] = {'disease_family': str_split(is_a),
                                       'disease_synonym': disease_synonym_split,
                                       'disease_definition': defn}

                # Update `ont_disease_synonym_dict`
                if isinstance(disease_synonym_split, tuple):
                    for s in disease_synonym_split:
                        if s not in ont_disease_synonym_dict:
                            ont_disease_synonym_dict[s] = [name]
                        # Check a duplicate is not added under a given disease_synonym
                        elif name not in ont_disease_synonym_dict[s]:
                            ont_disease_synonym_dict[s] += [name]

        return ont_name_dict, ont_name_dict_nest_keys, {k: sorted(v) for k, v in ont_disease_synonym_dict.items()}

    def __init__(self, cache_path=None, verbose=True):
        self.verbose = verbose
        # Load the database
        ontology_df = DiseaseOntInterface(cache_path=cache_path, verbose=verbose).pull()

        # Obtain dictionaries
        self.ont_name_dict, self.ont_name_dict_nest_keys, self.ont_disease_synonym_dict = self._dis_ont_dict_gen(
            ontology_df)

        # Convert `ont_name_dict_nest_keys` to an empty dict.
        self.empty_nest_dict = dict.fromkeys(self.ont_name_dict_nest_keys, np.NaN)

        # Extract keys from the two dictionaries passed
        self.ont_name_dict_keys = tuple(self.ont_name_dict.keys())
        self.ont_disease_synonym_dict_keys = tuple(self.ont_disease_synonym_dict.keys())

    def _disease_synonym_match(self, disease_synonym):
        """

        Maps a disease synonym to an *actual* disease name.
        The 'disease_synonym' key of the dictionary which is returned will have `disease_synonym` removed
        and will have a disease names which are mapped to `disease_synonym` installed in its place.

        Put another way, `ont_name_dict` gives the formal name. If we have a disease which is not in
        this dictionary, we may find it in a list of synonyms associated with that disease.

        :param disease_synonym: a disease synonym.
        :param disease_synonym: ``str``
        :return: data for a disease which the input `disease_synonym` is a synonym.
        :rtype: ``dict``
        """
        # Mapping from disease_synonym to related diseases
        ont_dis_names = self.ont_disease_synonym_dict[disease_synonym]

        # Simply use the first disease name related to the disease_synonym.
        # Note: this *assumes* that which 'name' is chosen from the list is irrelevant.
        # If the disease ontology database is not consistent, this assumption is invalid.
        disease_info = deepcopy(self.ont_name_dict[ont_dis_names[0]])
        # Remove the synonym from the 'disease_synonym' key and add 'ont_dis_names'
        if isinstance(disease_info['disease_synonym'], tuple):
            disease_synonym_new = [i for i in disease_info['disease_synonym'] if i != disease_synonym] + ont_dis_names
        else:
            disease_synonym_new = [ont_dis_names]

        # Add to `disease_info` (and remove any possible duplicates)
        disease_info['disease_synonym'] = tuple(sorted(set(disease_synonym_new)))

        return disease_info

    def _find_disease_info_raw(self, disease):
        """

        Try to match the input (`disease`) to information in the Disease Ontology Database.

        :param disease: a disease name.
        :param disease: ``str``
        :return: information on the disease (see ``_dis_ont_dict_gen()``).
        :rtype: ``dict`` or ``None``
        """
        if not isinstance(disease, str):
            return None
        elif disease in self.ont_name_dict:
            return self.ont_name_dict[disease]
        elif disease in self.ont_disease_synonym_dict:
            return self._disease_synonym_match(disease_synonym=disease)
        else:
            return None

    def _find_disease_info(self, disease, fuzzy_threshold):
        """

        Look up the family, synonyms and definition for a given ``disease``.

        :param disease: a disease name.
        :param disease: ``str``
        :param fuzzy_threshold: an integer on ``(0, 100]``.
        :type fuzzy_threshold: ``int``, `bool`, ``None``
        :return: disease information dictionary.
        :rtype: ``dict``
        """
        # ToDo: add memorizing of fuzzy matches
        # Try matching the string raw (i.e., 'as is').
        raw_rslt = self._find_disease_info_raw(disease)
        if isinstance(raw_rslt, dict):
            return raw_rslt

        if is_int(fuzzy_threshold):
            process = try_fuzzywuzzy_import()

        # Eject if fuzzy matching is disabled
        if not is_int(fuzzy_threshold):
            return self.empty_nest_dict

        # Try using `ont_name_dict`
        name_fuzzy_match, threshold = process.extractOne(disease, self.ont_name_dict_keys)
        if threshold >= fuzzy_threshold:
            return self.ont_name_dict[name_fuzzy_match]

        # Try using `ont_disease_synonym_dict`
        disease_synonym_fuzzy_match, threshold = process.extractOne(disease, self.ont_disease_synonym_dict_keys)
        if threshold >= fuzzy_threshold:
            return self._disease_synonym_match(disease_synonym_fuzzy_match)
        else:
            return self.empty_nest_dict

    def integration(self, data_frame, fuzzy_threshold=False):
        """

        Create the 'disease_family', 'disease_synonym' and 'disease_definition' columns to ``data_frame``
        using Disease Ontology data.

        :param data_frame: a dataframe which has been passed through ``_ImagesInterfaceIntegration().integration()``
        :type data_frame: ``Pandas DataFrame``
        :param fuzzy_threshold: an integer on ``(0, 100]``.
        :type fuzzy_threshold: ``int``, `bool`, ``None``
        :return: ``data_frame`` with the columns enumerated in the description.
        :rtype: ``Pandas DataFrame``
        """
        if fuzzy_threshold is True:
            raise ValueError("`fuzzy_threshold` cannot be `True`. Please provide a specific integer on ``(0, 100]``.")

        # Extract disease information using the Disease Ontology database
        disease_ontology_data = [self._find_disease_info(i, fuzzy_threshold)
                                 for i in tqdm(data_frame['disease'], desc='Disease Data', disable=not self.verbose)]

        # Convert `disease_ontology_data` to a dataframe
        disease_ontology_addition = pd.DataFrame(disease_ontology_data)

        # Add the columns in `disease_ontology_addition` to `data_frame`.
        for c in self.ont_name_dict_nest_keys:
            data_frame[c] = disease_ontology_addition[c]

        return data_frame


# ----------------------------------------------------------------------------------------------------------
# Tools to Add Data by Matching Against Disease and Disease disease_synonyms.
# ----------------------------------------------------------------------------------------------------------


def _disease_synonym_match_battery(disease, disease_synonyms, resource_dict, fuzzy_threshold):
    """

    Try to match ``disease`` and ``disease_synonyms`` in ``resource_dict``
    and return the corresponding nested dictionary (i.e., value).

    :param disease: a disease name.
    :param disease: ``str``
    :param disease_synonyms: synonyms for ``disease``
    :type disease_synonyms: ``tuple``
    :param resource_dict: a nested dictionary (see: ``_DiseaseOntologyIntegration()._dis_ont_dict_gen()``).
    :type resource_dict: ``dict``
    :param fuzzy_threshold: an integer on ``(0, 100]``.
    :type fuzzy_threshold: ``int``, `bool`, ``None``
    :return: the nested dictionary for a given key.
    :rtype: ``dict`` or ``None``
    """
    # Extract the keys
    lookup_dict_keys = tuple(resource_dict.keys())

    # Import process
    if is_int(fuzzy_threshold):
        process = try_fuzzywuzzy_import()

    # Try disease 'as is'
    if disease in resource_dict:
        return resource_dict[disease]

    # Search through disease_synonyms
    if isinstance(disease_synonyms, tuple):
        for s in disease_synonyms:
            if s in resource_dict:
                return resource_dict[s]

    # Eject if fuzzy matching is disabled
    if not is_int(fuzzy_threshold):
        return np.NaN

    # Try Fuzzy matching on `disease`
    disease_fuzzy_match, threshold = process.extractOne(disease, lookup_dict_keys)
    if threshold >= fuzzy_threshold:
        return resource_dict[disease_fuzzy_match]

    # Try Fuzzy matching on `disease_synonyms`
    if not isinstance(disease_synonyms, tuple):
        return np.NaN
    else:
        for s in disease_synonyms:
            disease_synonym_fuzzy_match, threshold = process.extractOne(s, lookup_dict_keys)
            if threshold >= fuzzy_threshold:
                return resource_dict[disease_synonym_fuzzy_match]
        else:
            return np.NaN  # capitulate


def _resource_integration(data_frame, resource_dict, fuzzy_threshold, new_column_name, verbose, desc):
    """

    Integrates information in ``resource_dict`` into ``data_frame`` as new column (``new_column_name``).

    :param data_frame: a dataframe which has been passed through ``_DiseaseOntologyIntegration().integration()``
    :type data_frame: ``Pandas DataFrame``
    :param resource_dict: a nested dictionary (see: ``_DiseaseOntologyIntegration()._dis_ont_dict_gen()``).
    :type resource_dict: ``dict``
    :param fuzzy_threshold: an integer on ``(0, 100]``.
    :type fuzzy_threshold: ``int``, `bool`, ``None``
    :param new_column_name: the name of the column with the extracted information.
    :type new_column_name: ``str``
    :param verbose: If ``True``, print notice when downloading database.
    :type verbose: ``bool``
    :param desc: description to pass to ``tqdm``.
    :type desc: ``str`` or ``None``
    :return: ``data_frame`` with information extracted from ``resource_dict``
    :rtype: ``Pandas DataFrame``
    """
    if fuzzy_threshold is True:
        raise ValueError("`fuzzy_threshold` cannot be `True`. Please specify a specific integer on ``(0, 100]``.")

    missing_column_error_message = "`data_frame` must contain a '{0}' column.\n" \
                                   "Call ``_DiseaseOntologyIntegration().disease_ont_integration()``"

    if 'disease' not in data_frame.columns:
        raise AttributeError(missing_column_error_message.format('disease'))
    elif 'disease_synonym' not in data_frame.columns:
        raise AttributeError(missing_column_error_message.format('disease_synonym'))

    # Map gene-disease information onto the dataframe
    matches = list()
    for _, row in tqdm(data_frame.iterrows(), total=len(data_frame), desc=desc, disable=not verbose):
        match = _disease_synonym_match_battery(disease=row['disease'],
                                               disease_synonyms=row['disease_synonym'],
                                               resource_dict=resource_dict,
                                               fuzzy_threshold=fuzzy_threshold)
        matches.append(match)

    # Add the `rslt` series to `data_frame`
    data_frame[new_column_name] = matches

    return data_frame


# ----------------------------------------------------------------------------------------------------------
# Disease Symptoms Interface (Symptomatology)
# ----------------------------------------------------------------------------------------------------------


class _DiseaseSymptomsIntegration(object):
    """

    Integration of Disease Symptoms information.

    :param cache_path: location of the BioVida cache. If one does not exist in this location, one will created.
    Default to ``None`` (which will generate a cache in the home folder).
    :type cache_path: ``str`` or ``None``
    :param verbose: If ``True``, print notice when downloading database. Defaults to ``True``.
    :type verbose: ``bool``
    """

    @staticmethod
    def _disease_symptom_dict_gen(dis_symp_db):
        """

        Tool to create a dictionary mapping disease to symptoms.

        :param dis_symp_db: yield of ``DiseaseSymptomsInterface().pull()``
        :type dis_symp_db: ``Pandas DataFrame``
        :return: a dictionary of the form ``{disease name: [symptom, symptom, symptom, ...], ...}``
        :rtype: ``dict``
        """
        d = defaultdict(set)
        for disease, symptom in zip(dis_symp_db['common_disease_name'], dis_symp_db['common_symptom_term']):
            d[disease.lower()].add(symptom.lower())
        return {k: tuple(sorted(v)) for k, v in d.items()}

    def __init__(self, cache_path=None, verbose=True):
        self.verbose = verbose
        # Load the Disease Symptoms database
        dis_symp_db = DiseaseSymptomsInterface(cache_path=cache_path, verbose=verbose).pull()

        # Create a disease-symptoms mapping
        self.disease_symptom_dict = self._disease_symptom_dict_gen(dis_symp_db)

    def _mentioned_symptoms(self, data_frame):
        """

        Match 'known_associated_symptoms' to the 'abstract' for the given row.

        :param data_frame: ``updated_data_frame`` as evolved in ``_DiseaseSymptomsIntegration().integration()``.
        :type data_frame: ``Pandas DataFrame``
        :return: a series with tuples of 'known_associated_symptoms' found in 'abstract'.
        :rtype: ``Pandas Series``
        """
        # ToDo: consider using 'mesh' and 'problems' cols - would have to be added in ``_ImagesInterfaceIntegration()``.
        def match_symptoms(row):
            """Find items in 'known_associated_symptoms' in 'abstract'."""
            if isinstance(row['known_associated_symptoms'], (list, tuple)) and isinstance(row['abstract'], str):
                abstract_lower = row['abstract'].lower()
                symptoms = [i for i in row['known_associated_symptoms'] if i in abstract_lower]
                return tuple(symptoms) if len(symptoms) else np.NaN
            else:
                return np.NaN

        return [match_symptoms(row) for _, row in tqdm(data_frame.iterrows(),
                                                       total=len(data_frame),
                                                       desc='Matching Symptoms',
                                                       disable=not self.verbose)]

    def integration(self, data_frame, fuzzy_threshold=False):
        """

        Adds a 'known_associated_symptoms' column to ``data_frame`` based on the Disease Symptoms database.

        :param data_frame: a dataframe which has been passed through ``_DiseaseOntologyIntegration().integration()``.
        :type data_frame: ``Pandas DataFrame``
        :param fuzzy_threshold: an integer on ``(0, 100]``.
        :type fuzzy_threshold: ``int``, ``bool``, ``None``
        :return: ``data_frame`` with a 'known_associated_symptoms' column.
        :rtype: ``Pandas DataFrame``
        """
        # Generate a 'known_associated_symptoms' columns
        updated_data_frame = _resource_integration(data_frame=data_frame,
                                                   resource_dict=self.disease_symptom_dict,
                                                   fuzzy_threshold=fuzzy_threshold,
                                                   new_column_name='known_associated_symptoms',
                                                   verbose=self.verbose,
                                                   desc='Symptoms')

        # Find 'known_associated_symptoms' which individual patients presented with by scanning the abstract
        updated_data_frame['mentioned_symptoms'] = self._mentioned_symptoms(updated_data_frame)

        # Drop the 'abstract' column as it is no longer needed
        del updated_data_frame['abstract']

        return updated_data_frame


# ----------------------------------------------------------------------------------------------------------
# DisGeNET Integration
# ----------------------------------------------------------------------------------------------------------


class _DisgenetIntegration(object):
    """

    Integration of DisGeNET information.

    :param cache_path: location of the BioVida cache. If one does not exist in this location, one will created.
    Default to ``None`` (which will generate a cache in the home folder).
    :type cache_path: ``str`` or ``None``
    :param verbose: If ``True``, print notice when downloading database. Defaults to ``True``.
    :type verbose: ``bool``
    """

    @staticmethod
    def _disease_gene_dict_gen(disgenet_df):
        """

        Generates a dictionary of the form: ``{disease name: (gene name, disgenet score), ...}``

        :param disgenet_df: yield of ``DisgenetInterface().pull('all)``.
        :type disgenet_df: ``Pandas DataFrame``
        :return: dictionary of the form ``{'disease_name': [('gene_name', disgenet score), ...], ...}``.
        :rtype: ``dict``
        """
        d = defaultdict(list)
        cols = zip(*[disgenet_df[c] for c in ('disease_name', 'gene_name', 'score')])
        for disease_name, gene_name, score in cols:
            d[disease_name].append((gene_name, score))
        return {k: tuple(sorted(v, key=lambda x: x[0])) for k, v in d.items()}

    def __init__(self, cache_path=None, verbose=True):
        self.verbose = verbose
        # Load the database
        disgenet_df = DisgenetInterface(cache_path=cache_path, verbose=verbose).pull('all')

        # Extract the relevant information in `disgenet_df` to a dictionary.
        self.disease_gene_dict = self._disease_gene_dict_gen(disgenet_df)

    def integration(self, data_frame, fuzzy_threshold=False):
        """

        Adds a series of genes known to be associated with the given disease to ``data_frame``.

        :param data_frame: a dataframe which has been passed through ``_DiseaseOntologyIntegration().integration()``
        :type data_frame: ``Pandas DataFrame``
        :param fuzzy_threshold: an integer on ``(0, 100]``.
        :type fuzzy_threshold: ``int``, `bool`, ``None``
        :return: ``data_frame`` with a 'known_associated_genes' column.
        :rtype: ``Pandas DataFrame``
        """
        return _resource_integration(data_frame=data_frame,
                                     resource_dict=self.disease_gene_dict,
                                     fuzzy_threshold=fuzzy_threshold,
                                     new_column_name='known_associated_genes',
                                     verbose=self.verbose,
                                     desc='Genomic')


# ----------------------------------------------------------------------------------------------------------
# Unify
# ----------------------------------------------------------------------------------------------------------


def images_unify(instances, db_to_extract='records_db', verbose=True, fuzzy_threshold=False):
    """

    Unify Instances in the ``images`` subpackage against other BioVida APIs.

    :param instances: See: ``biovida.unify_domains.unify_against_images()``
    :param instances: `list``, ``tuple``, ``OpeniInterface``, ``OpeniImageProcessing`` or ``CancerImageInterface``
    :param db_to_extract: the database to use. Must be one of: 'records_db', 'cache_records_db'.
                          Defaults to 'records_db'.
    :type db_to_extract: ``str``
    :param verbose: See: ``biovida.unify_domains.unify_against_images()``
    :param verbose: ``bool``
    :param fuzzy_threshold: See: ``biovida.unify_domains.unify_against_images()``
    :param fuzzy_threshold: ``int``, ``bool``, ``None``
    :return: See: ``biovida.unify_domains.unify_against_images()``
    :rtype: ``Pandas DataFrame``
    """
    instances = [instances] if not isinstance(instances, (list, tuple)) else instances

    # Catch ``fuzzy_threshold=True`` and set to a reasonably high default.
    if fuzzy_threshold is True:
        fuzzy_threshold = 95

    # Note: this doesn't consider cases where multiple biovida
    # caches exist, each with different data, e.g., one has
    # genetic data, another has Symptoms data. Thus, in rare
    # cases, this may cause data to be downloaded unnecessarily.
    cache_path = None  # default
    for i in instances:
        if hasattr(i, '_cache_path'):
            cache_path = getattr(i, '_cache_path')
            break

    # Combine Instances
    combined_df = _ImagesInterfaceIntegration().integration(instances=instances, db_to_extract=db_to_extract)

    # Disease Ontology
    combined_df = _DiseaseOntologyIntegration(cache_path, verbose).integration(combined_df, fuzzy_threshold)

    # Disease Symptoms
    combined_df = _DiseaseSymptomsIntegration(cache_path, verbose).integration(combined_df, fuzzy_threshold)

    # Disgenet
    combined_df = _DisgenetIntegration(cache_path, verbose).integration(combined_df, fuzzy_threshold)

    return combined_df



# coding: utf-8

"""

    Processing the 'image_id' Column Returned by Open-i
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

"""
import re
from collections import defaultdict

# General Support Tools
from biovida.support_tools.support_tools import cln, only_numeric, multi_replace


_image_id_cleaner_regex = re.compile('\.[A-Za-z]+$')
_image_id_medpix_regex = re.compile('\d+$')
_image_id_trail_regex = re.compile('\d+-\d+(?:[a-z])?$')
_image_id_zero_regex = re.compile('^0+')


def _image_id_parser(image_id):
    """

    Extract the following pattern from ``image_id``:
    'f...numbers(optional: one letter from a-g)'.

    :param image_id: an element from the 'image_id' column.
    :type image_id: ``str``
    :return: see description and examples in ``_image_id_short()``.
    :rtype: ``str`` or ``None``
    """
    allowed_letters = list('abcdefghij')
    if 'f' in image_id:
        look_for = 'f'
    elif 'g' in image_id:
        look_for = 'g'
    else:
        return None

    candidate = ''
    do_add = False
    hit_digit = False
    letters_added = False
    drop_letters = False
    for c in image_id.lower():
        if c == look_for:
            do_add = True
            continue
        if do_add and c.isdigit():
            candidate += c
            hit_digit = True
        elif hit_digit and not letters_added and c in allowed_letters:
            candidate += c
            letters_added = True
            continue
        if hit_digit and letters_added and c.isalpha():
            drop_letters = True
        if hit_digit and not c.isdigit():
            break
    if len(candidate) and drop_letters:
        return only_numeric(candidate, float_output=False)
    elif len(candidate):
        return candidate
    else:
        return None


def _image_id_cleaner(image_id):
    """

    Performs the following changes
    to ``image_id``.

    1. File extensions
    2. Convert 'image' and 'img' to 'f'
    3. Clean white spaces and lower.

    :param image_id: an element from the 'image_id' column.
    :type image_id: ``str``
    :return: see description.
    :rtype: ``str``
    """
    image_id_no_ext = re.sub(_image_id_cleaner_regex, "", image_id)
    return cln(multi_replace(image_id_no_ext, ('image', 'img'), replace_with='f')).lower()


def _image_id_short(journal_title, image_id):
    """

    Simplify ``image_id`` to determine the figure number.

    :param journal_title: an element from the 'journal_title' column.
    :type journal_title: ``None`` or ``str``
    :param image_id: an element from the 'image_id' column.
    :type image_id: ``str``
    :return: a cleaned ``image_id``
    :rtype: ``str`` or ``None``

    :Example:

    >>> _image_id_short(journal_title='medpix', image_id='MPX..16962')
    ...
    '16962'
    >>> _image_id_short(journal_title='pubmed', image_id='f06')
    ...
    '6'
    >>> _image_id_short(journal_title='pubmed', image_id='f9')
    ...
    '9'
    >>> _image_id_short(journal_title='pubmed', image_id='new-img009-mri')
    ...
    '9'
    >>> _image_id_short(journal_title='pubmed', image_id='image0011')
    ...
    '11'
    >>> _image_id_short(journal_title='unknown', image_id='figure4a')
    ...
    '4a'
    >>> _image_id_short(journal_title='unknown', image_id='figure4az')
    ...
    '4'
    >>> _image_id_short(journal_title='pubmed', image_id='f1b-new')
    ...
    '1b'
    >>> _image_id_short(journal_title=None, image_id='f1bz')
    ...
    '1'
    >>> _image_id_short(journal_title='pubmed', image_id='f019c-2')
    ...
    '2'
    >>> _image_id_short(journal_title='pubmed', image_id='f319cg-33')
    ...
    '33'
    >>> _image_id_short(journal_title=np.NaN, image_id='f2b-993')
    ...
    '2b' # not confused by the trailing '999' as it is >> 2, and thus less likely to be the fig. no.
    """
    if not isinstance(image_id, str):
        return None

    image_id_clean = _image_id_cleaner(image_id)

    if isinstance(journal_title, str) and 'medpix' in journal_title.lower():
        medpix_match = re.findall(_image_id_medpix_regex, image_id_clean)
        if len(medpix_match) == 1 and len(medpix_match[0]):
            return medpix_match[0]

    end = re.findall(_image_id_trail_regex, image_id_clean)
    end_match = end[0].replace("-", "") if len(end) == 1 and len(end[0]) else None
    p_image_id = _image_id_parser(image_id_clean)

    def output_decision(a, b):
        small_ceiling = 100
        to_compare = list()
        for i in (a, b):
            try:
                to_compare.append([i, only_numeric(i)])
            except:
                pass
        if len(to_compare) != 2:
            return None
        if any(i[1] < small_ceiling for i in to_compare):
            return min(to_compare, key=lambda x: x[1])[0]
        else:
            return max(to_compare, key=lambda x: x[1])[0]

    def output_clean(output):
        if len([i for i in output if i.isdigit() and i != '0']):
            return cln(re.sub(_image_id_zero_regex, '', output))
        else:
            return cln(output)

    if end_match is None and p_image_id is None:
        return None
    elif isinstance(end_match, str) and isinstance(p_image_id, str):
        fig_no = output_decision(end_match, p_image_id)
    elif isinstance(end_match, str):
        fig_no = end_match
    elif isinstance(p_image_id, str):
        fig_no = p_image_id
    return output_clean(fig_no)


def _image_id_short_enforce_unique(data_frame):
    """

    If the 'image_id_short' column is not unique
    with respect to 'uid', make it unique by adding a
    an integer.

    :param data_frame: as evolved inside ``image_id_short_gen()``
    :type data_frame: ``Pandas DataFrame``
    :return: see description
    :rtype: ``Pandas DataFrame``


    >>> print(data_frame)
    ...
                 uid             image_id        image_id_short
        1    PMC4372763  sap-27-01-052-g002.tif         None
        2   PMC4593917              FI0192cr-1          192
        3   PMC4593917              FI0192cr-2          192
    ...
    >>> print(_image_id_short_enforce_unique(data_frame))
    ...
                  uid             image_id        image_id_short
        1    PMC4372763  sap-27-01-052-g002.tif           1
        2   PMC4593917              FI0192cr-1        192_1
        3   PMC4593917              FI0192cr-2        192_2

    """
    def not_unique_image_id_short(row):
        """Check if elements in the 'image_id_short' series are unique."""
        image_id_short = row['image_id_short'].tolist()
        if any(not isinstance(i, str) for i in image_id_short):
            return True
        return not len(set(image_id_short)) == len(image_id_short)

    # Dict of the form: {'uid': not unique or unique (boolean), ...}
    invalid_si_id_dict = data_frame.groupby('uid').apply(not_unique_image_id_short).to_dict()

    d = defaultdict(int)
    # Iterate over only those rows where 'image_id_short' is not unique and correct accordingly.
    for index, row in data_frame.iterrows():
        if invalid_si_id_dict.get(row['uid']):
            d[row['uid']] += 1
            if isinstance(row['image_id_short'], str):
                new_name = "{0}_{1}".format(row['image_id_short'],str(d[row['uid']]))
                data_frame.set_value(index, 'image_id_short', new_name)
            else:
                data_frame.set_value(index, 'image_id_short', str(d[row['uid']]))
    return data_frame


def image_id_short_gen(data_frame):
    """

    Simplify 'image_id' by 'boiling it down'
    to the figure number.

    Note: this will mutate ``data_frame`` in memory.

    :param data_frame: an Openi ``records_db`` or ``cache_records_db``.
    :type data_frame: ``Pandas DataFrame``
    :return: ``data_frame`` with a ``image_id_short`` column.
    :rtype: ``Pandas DataFrame``
    """
    data_frame['image_id_short'] = data_frame.apply(
        lambda x: _image_id_short(x['journal_title'], x['image_id']), axis=1)

    return _image_id_short_enforce_unique(data_frame)



# coding: utf-8

"""

    Printing Tools
    ~~~~~~~~~~~~~~

"""
import pandas as pd
from copy import deepcopy

# General Support Tools
from biovida.support_tools.support_tools import cln, pstr, items_null

# Suppress Pandas' SettingWithCopyWarning
pd.options.mode.chained_assignment = None


# ---------------------------------------------------------------------------------------------
# Dictionary Printing Suite
# ---------------------------------------------------------------------------------------------


def _key_padding(s, len_longest_key):
    """

    Add padding a key in a dict. for `dict_pretty_printer()`.

    :param s:
    :param len_longest_key:
    :return:
    """
    if not isinstance(s, str):
        return s
    return "{0}:{1}".format(cln(s).replace("_", " ").title(), " " * abs(len_longest_key - len(s)))


def _value_padding(s, len_longest_key, print_mold):
    """

    Add padding a value in a dict. for `dict_pretty_printer()`.

    :param s:
    :param len_longest_key:
    :return:
    """
    if not isinstance(s, str):
        return s
    padding = len_longest_key + (len(print_mold) - 3) + 2  # 2 = 1 space for the colon + 1 space added by `print()`.
    return s.replace("\n", "\n{0}".format(" " * padding))


def _char_in_braces(full_s, char_position):
    """

    Checks if a given position in a string lies between braces.
    Note: This is a rough solution, but more than sufficient. A robust solution would require a state machine.

    :param full_s: a string.
    :type full_s: ``str``
    :param char_position: a position in the string.
    :type char_position: ``int``
    :return: ``True`` if yes, else ``False``
    :rtype: bool
    """
    if not isinstance(full_s, str):
        raise ValueError('`full_s` must be a string.')

    if not isinstance(char_position, int):
        raise ValueError('`char_position` must be an int.')

    for pair in ('()', '[]', '{}'):
        if pair[0] in full_s[:char_position] and pair[1] in full_s[char_position:]:
            return True
    else:
        return False


def _value_correction(s, len_longest_key, max_value_length, print_mold):
    """

    Formats a value in a dict. for `dict_pretty_printer()`.

    :param s:
    :param len_longest_key:
    :param max_value_length:
    :param print_mold:
    :return:
    """
    if not isinstance(s, str):
        return s

    # Clean the input
    s_cleaned = cln(s).replace("\n", " ")

    # Get the position of all spaces in the cleaned string -- spaces inside braces are excluded.
    spaces = [i for i, c in enumerate(s_cleaned) if c == " " and not _char_in_braces(s_cleaned, i)]

    # Return if the string is shorter than the `max_value_length` or there are no space.
    if len(s_cleaned) < max_value_length or not len(spaces):
        return _value_padding(s_cleaned, len_longest_key, print_mold)

    # Identify ideal points for a line break, w.r.t. max_value_length.
    ideal_break_points = [i for i in range(len(s_cleaned)) if i % max_value_length == 0 and i != 0]

    # Get the possible break points
    true_break_points = [min(spaces, key=lambda x: abs(x - ideal)) for ideal in ideal_break_points]

    if len(true_break_points):
        formatted_string = "".join([c if e not in true_break_points else "\n" for e, c in enumerate(s_cleaned)])
    else:
        formatted_string = s_cleaned

    return _value_padding(formatted_string, len_longest_key, print_mold)


def dict_pprint(d, max_value_length=70):
    """

    Pretty prints a dictionary with vertically aligned values.
    Dictionary values with strings longer than `max_value_length`
    are automatically broken and aligned with the line(s) above.

    :param d: a dictionary
    :type d: ``dict``
    :param max_value_length: max. number of characters in a string before a line break.
                            This is a fuzzy threshold because the algorithm will only insert
                            line breaks where there are already spaces and will not insert line breaks
                            between braces.
    :type max_value_length: ``int``

    :Example:

    >>> d = {
    'part_2': 'Lorem ipsum dolor sit amet, consectetur adipiscing elit. Ut dictum nisi sed euismod consequat. '
              'Donec tempus enim nec lorem ornare, non sagittis dolor consequat. Sed facilisis tortor '
              'vel enim mattis, et fermentum mi posuere.',
    'part_1': 'Duis sit amet nulla fermentum, vestibulum mauris at, porttitor massa. '
              'Vestibulum luctus interdum mattis.'
    }
    >>> dict_pprint(d)
     - Part 2:  Lorem ipsum dolor sit amet, consectetur adipiscing elit. Ut dictum nisi
                sed euismod consequat. Donec tempus enim nec lorem ornare, non sagittis
                dolor consequat. Sed facilisis tortor vel enim mattis, et fermentum
                mi posuere.
     - Part 1:  Duis sit amet nulla fermentum, vestibulum mauris at, porttitor massa.
                Vestibulum luctus interdum mattis.

    """
    print_mold = " - {0} "

    # Compute the length of the longest key
    len_longest_key = len(max(list(d.keys()), key=len))

    new_dict = {_key_padding(k, len_longest_key): _value_correction(v, len_longest_key, max_value_length, print_mold)
                for k, v in deepcopy(d).items()}

    # Print the dict
    for k, v in new_dict.items():
        print(print_mold.format(k), v)


# ----------------------------------------------------------------------------------------------------------
# Pandas Printing Suite
# ----------------------------------------------------------------------------------------------------------


def _padding(s, amount, justify):
    """

    Add padding to a string.

    :param s: a string.
    :type s: ``str``
    :param amount: the amount of white space to add.
    :type amount: ``float`` or ``int``
    :param justify: the justification of the resultant text. Must be one of: 'left' or 'center'.
    :type justify: ``str``
    :return: `s` justified, or as passed if `justify` is not one of: 'left' or 'center'.
    :rtype: ``str``
    """
    # Source: https://github.com/TariqAHassan/EasyMoney
    pad = ' ' * amount
    if justify == 'left':
        return "%s%s" % (pstr(s), pad)
    elif justify == 'center':
        return "%s%s%s" % (pad[:int(amount/2)], pstr(s), pad[int(amount/2):])
    else:
        return s


def _pandas_series_alignment(pandas_series, justify):
    """

    Align all items in a pandas series.

    :param pandas_series: a pandas series.
    :type pandas_series: ``Pandas Series``
    :param justify: the justification of the resultant text. Must be one of: 'left', 'right' or 'center'.
    :type justify: ``str``
    :return: aligned series
    :rtype: ``str``
    """
    # Source: https://github.com/TariqAHassan/EasyMoney
    if justify == 'right':
        return pandas_series

    def series_to_string(x):
        if items_null(x):
            return x
        if isinstance(x, list) or 'ndarray' in str(type(x)):
            return "[{0}]".format(", ".join(map(str, list(x))))
        elif isinstance(x, tuple):
            pattern = "({0},)" if len(x) == 1 else "({0})"
            return pattern.format(", ".join(map(str, x)))
        elif 'pandas' in str(type(x)) and 'Timestamp' in str(type(x)):
            if all((x.hour == 0, x.minute == 0, x.second == 0)):
                return x.strftime('%Y-%m-%d')
            else:
                return str(x)
        else:
            return str(x)

    pandas_series_str = pandas_series.map(series_to_string)
    longest_string = max([len(s) for s in pandas_series_str.astype('unicode')])
    return [_padding(s, longest_string - len(s), justify) if not items_null(s) else s for s in pandas_series_str]


def _align_pandas(data_frame, to_align='right'):
    """

    Align the columns of a Pandas DataFrame by adding whitespace.

    :param data_frame: a dataframe.
    :type data_frame: ``Pandas DataFrame``
    :param to_align: 'left', 'right', 'center' or a dictionary of the form: ``{'Column': 'Alignment'}``.
    :type to_align: ``str``
    :return: dataframe with aligned columns.
    :rtype: ``Pandas DataFrame``
    """
    # Source: https://github.com/TariqAHassan/EasyMoney
    if isinstance(to_align, dict):
        alignment_dict = to_align
    elif to_align.lower() in ('left', 'right', 'center'):
        alignment_dict = dict.fromkeys(data_frame.columns, to_align.lower())
    else:
        raise ValueError("to_align must be either 'left', 'right', 'center', or a dict.")

    for col, justification in alignment_dict.items():
        data_frame[col] = _pandas_series_alignment(data_frame[col], justification)

    return data_frame


def _pandas_print_full(pd_df, full_rows, full_cols, suppress_index, column_width_limit):
    """

    Print *all* of a Pandas DataFrame.

    :param pd_df: DataFrame to printed in its entirety.
    :type pd_df: ``Pandas DataFrame``
    :param full_rows: print all rows if ``True``. Defaults to ``False``.
    :type full_rows: ``bool``
    :param full_cols: print all columns side-by-side if True. Defaults to ``True``.
    :type full_cols: ``bool``
    :param suppress_index: see ``pandas_pprint()``.
    :type suppress_index: ``bool``
    :param column_width_limit: change limit on how wide columns can be. If ``None``, no change will be made.
                               Defaults to ``None``.
    :type column_width_limit: ``int`` or ``None``
    """
    # Source: https://github.com/TariqAHassan/EasyMoney
    if full_rows:
        pd.set_option('display.max_rows', pd_df.shape[0])
    if full_cols:
        pd.set_option('expand_frame_repr', False)
        pd.set_option('display.max_columns', pd_df.shape[1])
    if not isinstance(column_width_limit, bool) and isinstance(column_width_limit, int):
        pd.set_option('display.width', column_width_limit)
        pd.set_option('display.max_colwidth', column_width_limit)

    if suppress_index:
        pd_df.index = [''] * pd_df.shape[0]

    print(pd_df)

    # Restore Pandas Printing Defaults
    if full_rows:
        pd.reset_option('display.max_rows')
    if full_cols:
        pd.set_option('expand_frame_repr', True)
        pd.set_option('display.max_columns', 20)
    if not isinstance(column_width_limit, bool) and isinstance(column_width_limit, int):
        pd.reset_option('display.width')
        pd.reset_option('display.max_colwidth')


def pandas_pprint(data,
                  col_align='right',
                  header_align='center',
                  full_rows=False,
                  full_cols=False,
                  suppress_index=False,
                  column_width_limit=None):
    """

    Pretty Print a Pandas DataFrame or Series.

    :param data: a dataframe or series.
    :type data: ``Pandas DataFrame`` or ``Pandas Series``
    :param col_align: 'left', 'right', 'center' or a dictionary of the form: ``{'column_name': 'alignment'}``,
                       e.g., ``{'my_column': 'left', 'my_column2': 'right'}``.
    :type col_align: ``str`` or ``dict``
    :param header_align: alignment of headers. Must be one of: 'left', 'right', 'center'.
    :type header_align: ``str`` or ``dict``
    :param full_rows: print all rows.
    :type full_rows: ``bool``
    :param full_cols: print all columns.
    :type full_cols: ``bool``
    :param suppress_index: if ``True``, suppress the index. Defaults to ``False``.
    :type suppress_index: ``bool``
    :param column_width_limit: change limit on how wide columns can be. If ``None``, no change will be made.
                               Defaults to ``None``.
    :type column_width_limit: ``int`` or ``None``
    """
    # Source: https://github.com/TariqAHassan/EasyMoney

    # ToDO: if col_align != 'right' and `full_rows=False`,
    # only the first 30 and last 30 rows need to be adjusted.

    if type(data).__name__ not in ('DataFrame', 'Series'):
        raise TypeError("`data` cannot be of type '{0}'; "
                        "must be of type 'DataFrame' or 'Series'.".format(type(data).__name__))

    # Deep copy to prevent altering ``data`` in memory.
    data_copy = data.copy(deep=True)

    if isinstance(data_copy, pd.Series):
        data_copy = data_copy.to_frame()

    aligned_df = _align_pandas(data_copy, col_align)
    pd.set_option('colheader_justify', header_align)
    _pandas_print_full(pd_df=aligned_df.fillna(""), full_rows=full_rows, full_cols=full_cols,
                       suppress_index=suppress_index, column_width_limit=column_width_limit)
    pd.set_option('colheader_justify', 'right')



# coding: utf-8

"""

    Unifying Across Domains
    ~~~~~~~~~~~~~~~~~~~~~~~

"""
# Subpackage Unification Tools -- Images
from biovida.images._unify_images_against_other_biovida_apis import images_unify


# ----------------------------------------------------------------------------------------------------------
# Unify Against Images
# ----------------------------------------------------------------------------------------------------------


def unify_against_images(instances,
                         db_to_extract='records_db',
                         verbose=True,
                         fuzzy_threshold=False):
    """

    Tool to unify image instances (namely ``OpeniInterface`` and/or ``CancerImageInterface``)
    with Diagnostic and Genomic Data.

    :param instances: any one of ``OpeniInterface``, ``CancerImageInterface`` or ``OpeniImageProcessing``, or some
                           combination inside an iterable.
    :type instances: ``list``, ``tuple``, ``OpeniInterface``, ``CancerImageInterface`` or ``OpeniImageProcessing``.
    :param db_to_extract: the database to use. Must be one of: 'records_db', 'cache_records_db' or 'image_dataframe'.
                          Defaults to 'records_db'.

                    .. note::

                        If an instance of ``OpeniImageProcessing`` is passed to ``instances``, the ``image_dataframe``
                        attribute will be extracted regardless of the value passed to this argument.

    :type db_to_extract: ``str``
    :param verbose: If ``True``, print notice when downloading database. Defaults to ``True``.
    :type verbose: ``bool``
    :param fuzzy_threshold: an integer on ``(0, 100]``. If ``True`` a threshold of `95` will be used. Defaults to ``False``.
    
                .. warning::

                        While this parameter will likely increase the number of matches, fuzzy searching with
                        large databases, such as those this function integrates, is very computationally expensive.
    
    :type fuzzy_threshold: ``int``, ``bool``, ``None``
    :return: a dataframe which unifies image instances with genomic and diagnostics data.
    :rtype: ``Pandas DataFrame``


    This function evolves a DataFrame with the following columns:

    .. hlist::
        :columns: 4

        * 'age'
        * 'article_type'
        * 'disease'
        * 'image_caption'
        * 'image_id'
        * 'modality_best_guess'
        * 'pull_time'
        * 'query'
        * 'sex'
        * 'source_api'
        * 'disease_family'
        * 'disease_synonym'
        * 'disease_definition'
        * 'known_associated_symptoms'
        * 'mentioned_symptoms'
        * 'known_associated_genes'


    .. note::

        The ``'known_associated_genes'`` column is of the form ``((Gene Name, DisGeNET Evidence Score), ...)``.

    .. note:: 
    
        If an instance of the ``CancerImageInterface`` class is passed, a ``'source_images_path'``
        column will also be generated if the database extracted from the instance contains 
        a ``'cached_dicom_images_path'`` column.

    .. warning::

        The ``'known_associated_symptoms'`` and ``'known_associated_genes'`` columns denote symptoms and genes
        known to be associated with the disease the patient presented with. **These columns are not an account
        of the symptomatology or genotype of the patients themselves**. Conversely, the ``'mentioned_symptoms'``
        column is an account of a given patient's symptoms *if* the data is from a clinical case
        (i.e., ``article_type`` equals 'case_report').

    :Example:

    >>> from biovida.images import OpeniInterface
    >>> from biovida.images import CancerImageInterface
    >>> from biovida.unification import unify_against_images
    ...
    >>> opi = OpeniInterface()
    # --- Search and Pull ---
    >>> udf1 = unify_against_images(opi)
    ...
    # Adding another Interface from the images subpackage
    >>> cii = CancerImageInterface(YOUR_API_KEY_HERE)
    # --- Search and Pull ---
    >>> udf2 = unify_against_images([opi, cii])
    """
    return images_unify(instances=instances, db_to_extract=db_to_extract,
                        verbose=verbose, fuzzy_threshold=fuzzy_threshold)



# coding: utf-8

"""

    Information for ``openi_text_processing._imaging_modality_guess()``
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

"""
from copy import deepcopy
from itertools import combinations


# Note: all terms for matching must be lower case (_imaging_modality_guess() lowers all strings).


# ----------------------------------------------------------------------------------------------------------
# Terms Dict
# ----------------------------------------------------------------------------------------------------------


terms_dict = {
    # format: {abbreviated name, ([alternative names], formal name)}
    # Note: formal names are intended to be collinear with `biovida.images._resources.openi_parameters`'s
    # `openi_image_type_modality_full` dictionary.
    "ct": (['ct ', 'ct:', 'ct-', ' ct', ' ct ', '(ct)',
            'computed tomography', '(ct)'], 'Computed Tomography (CT)'),
    "mri": (['mr ', ' mr ', 'mri ', 'mri:', 'mri-', ' mri', ' mri ', '(mri)',
             'magnetic resonance imag'], 'Magnetic Resonance Imaging (MRI)'),  # 'imag' catches 'es' and 'ing'.
    "pet": (['pet:', 'pet-', ' pet ', '(pet)',
             'positron emission tomography'], 'Positron Emission Tomography (PET)'),
    "photograph": (['photograph'], 'Photograph'),
    "ultrasound": (['ultrasound'], 'Ultrasound'),
    "mammogram": (['mammogram', 'mammograph'], 'Mammography'),
    "x-ray": (['xray', 'x-ray'], 'X-Ray')
}


# ----------------------------------------------------------------------------------------------------------
# Modality Specific Subtypes
# ----------------------------------------------------------------------------------------------------------


# Define a dictionary to use when the modality (term) itself is not present
# e.g., t2 is but MRI is not.
# Note: the first item in the lists below are the 'flag bearers'.
modality_specific_subtypes = {
    'ct': [
        ['angiography', 'angiogram', 'angio-ct'],
        ['chest', 'lung'],
        [' mpi '],
        ['segmentation'],
    ],
    'mri': [
        [' gadolinium ', ' gad '],
        ['post-gadolinium', 'post-gad', 'post gad ', ' post gad'],
        [' t1 ', 't1 ', 't1-', 't1w'],
        [' t2 ', 't2 ', 't2-', 't2w'],
        [' stir ', 'stir ', ' stir'],
        [' spgr ', 'spgr ', ' spgr'],
        [' fse ', 'fse ', ' fse '],
        ['flair'],
        [' mra ', 'mra ', ' mra'],
        [' dwi ', 'diffusion weighted'],
        [' dti ', 'diffusion tensor']
    ],
}


# ----------------------------------------------------------------------------------------------------------
# Modality Subtypes. Not specific & cannot be used to infer the modality (like 'T1', for example).
# ----------------------------------------------------------------------------------------------------------


# Divvy up terms which can be used if a match for ``terms_dict`` is successful,
# i.e., if the modality is known, these terms can be used to increase specificity.
contrast = ['contrast-enhanced', 'contrast enhanced', 'enhanced contrast', 'w/contrast', 'with contrast']

ct_mri_xray = [
    ['abdomen', 'abdominal'],
    ['chest', 'lung']
]

ct_and_mri = [
    ['brain', 'cranial', 'cranium'],
    ['spinal', 'spine'],
    ['non-contrast', 'non contrast', 'noncontrast', 'w/o contrast',
     'without contrast', 'unenhanced'],
    ['pre-contrast', 'precontrast'],
    ['post-contrast', 'postcontrast'],
]

# Define a dictionary to use when there is a match for an item in ``terms_dict``.
modality_subtypes = deepcopy(modality_specific_subtypes)
modality_subtypes['ct'] += ct_mri_xray + ct_and_mri + [contrast + [' contrast ct ', ' contrast-ct ']]
modality_subtypes['mri'] += ct_mri_xray + ct_and_mri + [contrast + [' contrast mr ', ' contrast-mr ']]
modality_subtypes['x-ray'] = ct_mri_xray


# ----------------------------------------------------------------------------------------------------------
# Contradictions
# ----------------------------------------------------------------------------------------------------------


# Add contradictions to block
contradictions = [
    ['t1', 't2'],
    ['contrast-enhanced', 'non-contrast'],
    ['pre-contrast', 'post-contrast'],
    ['gadolinium', 'post-gadolinium'],
]

# Add all possible contradictions generated by 'abdomen', 'chest', 'brain' and 'spinal'.
contradictions += list(map(list, combinations(['abdomen', 'brain', 'chest', 'spinal'], r=2)))



# coding: utf-8

"""

    Image Cache Management
    ~~~~~~~~~~~~~~~~~~~~~~

"""
import os
import shutil
import numpy as np
import pandas as pd
from warnings import warn
from collections import Counter
from collections import defaultdict

# General Support Tools
from biovida.support_tools.support_tools import (tqdm,
                                                 cln,
                                                 multimap,
                                                 directory_existence_handler,
                                                 ipython_display,
                                                 IN_NOTEBOOK)

# Utilities
from biovida.support_tools.utilities import (_tvt_dict_gen,
                                             train_val_test,
                                             _train_val_test_engine,
                                             _file_paths_dict_to_ndarrays)

# Import Printing Tools
from biovida.support_tools import pandas_pprint


# ----------------------------------------------------------------------------------------------------------
# Relationship Mapping Function
# ----------------------------------------------------------------------------------------------------------


def _openi_image_relation_map(data_frame):
    """

    Algorithm to find the index of rows which reference
    the same image in the cache (the image size can vary).

    :param data_frame:
    :type data_frame: ``Pandas DataFrame``
    :return:
    :rtype: ``Pandas DataFrame``
    """
    # Copy the data_frame
    df = data_frame.copy(deep=True)

    # Reset the index
    df = df.reset_index(drop=True)

    # Get duplicated img_large occurrences. Use of 'img_large' is arbitrary, could have used
    # any of the 'img_...' columns, e.g., 'img_thumb' or 'img_grid150'.
    duplicated_image_refs = (k for k, v in Counter(df['img_large']).items() if v > 1)

    # Get the indices of duplicates
    dup_index = {k: df[df['img_large'] == k].index.tolist() for k in duplicated_image_refs}

    def related(image_large, index):
        """Function to look for references to the same image in the cache."""
        if image_large in dup_index:
            return tuple(sorted([i for i in dup_index[image_large] if i != index]))
        else:
            return np.NaN

    # Apply `relate()`
    df['shared_image_ref'] = [related(image, index) for image, index in zip(df['img_large'], df.index)]

    return df


# ----------------------------------------------------------------------------------------------------------
# Merging New Records with the Existing Cache
# ----------------------------------------------------------------------------------------------------------


def _dict_to_tot(d):
    """

    Convert a dictionary to a tuple of tuples and sort by the former keys.

    :param d: any dictionary.
    :type d: ``dict``
    :return: ``d`` as a tuple of tuples, sorted by the former key values.
    :rtype: ``tuple``
    """
    if not isinstance(d, dict):
        return d

    def values_to_tuples(d):
        return {k: tuple(sorted(v)) if isinstance(v, (list, tuple)) else v for k, v in d.items()}

    return tuple(sorted(values_to_tuples(d).items(), key=lambda x: x[0]))


def _record_update_dbs_joiner(records_db, update_db):
    """

    Join and drop rows for which `update_db`'s columns exclusively contain NaNs.

    :param records_db: permanent database/dataframe which keeps a record of files in the cache.
    :type records_db: ``Pandas DataFrame``
    :param update_db: database/dataframe to 'update' ``records_db``
    :type update_db: ``Pandas DataFrame``
    :return: ``records_db`` with ``update_db`` left-joined.
    :rtype: ``Pandas DataFrame``
    """
    try:
        joined_db = records_db.join(update_db, how='left')
    except:
        joined_db = records_db
        for c in update_db.columns:
            joined_db[c] = update_db[c]

    return joined_db.fillna(np.NaN).dropna(subset=list(update_db.columns), how='all').reset_index(drop=True)


def _relationship_mapper(data_frame, interface_name):
    """

    :param data_frame:
    :type data_frame: ``Pandas DataFrame``
    :param interface_name:
    :type interface_name: ``str``
    :return:
    :rtype: ``Pandas DataFrame``
    """
    _relationship_mapping_dict = {
        # Keys: Interface Class Name.
        # Values: mapping function
        'OpeniInterface': _openi_image_relation_map
    }

    if interface_name in _relationship_mapping_dict:
        relationship_mapping_func = _relationship_mapping_dict[interface_name]
        data_frame = relationship_mapping_func(data_frame)

    return data_frame


def _records_db_merge(interface_name,
                      current_records_db,
                      records_db_update,
                      columns_with_dicts,
                      duplicates,
                      rows_to_conserve_func=None,
                      pre_return_func=None,
                      columns_with_iterables_to_sort=None):
    """

    Merge the existing cache record database with new additions.

    .. warning::

        Both ``current_records_db`` and ``records_db_update`` are
        expected to have 'biovida_version' and 'pull_time' columns.

    :param current_records_db: the existing record database.
    :type current_records_db: ``Pandas DataFrame``
    :param records_db_update: the new records dataframe to be merged with the existing one (``current_records_db``).
    :type records_db_update: ``Pandas DataFrame``
    :param columns_with_dicts: a list of columns which contain dictionaries. Note: this column *should* contain only
                               dictionaries or NaNs.
    :type columns_with_dicts: ``list``, ``tuple`` or ``None``
    :param duplicates: A function to handle dropping duplicates. This function should accept a dataframe
                      (and *only* a dataframe) and return a dataframe.
    :type duplicates: ``func``
    :param rows_to_conserve_func: function to generate a list of booleans which denote whether or not the image is,
                                  in fact, present in the cache. If not, remove it from the database to be saved.
    :type rows_to_conserve_func: ``func``
    :param pre_return_func: a function to apply to the dataframe before it is returned. This function should accept a
                            dataframe (and *only* a dataframe) and return a dataframe.
    :type pre_return_func: ``None`` or ``func``
    :param columns_with_iterables_to_sort: columns which themselves contain lists or tuples which should be sorted
                                           prior to dropping. Defaults to ``None``.
    :type columns_with_iterables_to_sort: ``list`` or ``tuple``
    :return: a dataframe which merges ``current_records_db`` and ``records_db_update``
    :rtype: ``Pandas DataFrame``
    """
    # Note: this function does not explicitly handle cases where
    #       combined_dbs has length 0, no obvious need to though.
    combined_dbs = pd.concat([current_records_db, records_db_update], ignore_index=True)

    # Mark each row to conserve order following ``pandas.drop_duplicates()``.
    combined_dbs['__temp_order__'] = range(combined_dbs.shape[0])

    if callable(rows_to_conserve_func):
        combined_dbs = combined_dbs[combined_dbs.apply(rows_to_conserve_func, axis=1)]

    # Convert items in ``columns_with_dicts`` from dictionaries to tuple of tuples.
    # (making them hashable, as required by ``pandas.drop_duplicates()``).
    combined_dbs = multimap(combined_dbs, columns=columns_with_dicts, func=_dict_to_tot)

    combined_dbs = multimap(combined_dbs, columns=columns_with_iterables_to_sort,
                            func=lambda x: tuple(sorted(x)))

    combined_dbs = duplicates(combined_dbs)
    combined_dbs = multimap(combined_dbs, columns=columns_with_dicts, func=dict)
    combined_dbs = combined_dbs.sort_values('__temp_order__')
    combined_dbs = _relationship_mapper(data_frame=combined_dbs, interface_name=interface_name)

    if callable(pre_return_func):
        combined_dbs = pre_return_func(combined_dbs)

    return combined_dbs.drop('__temp_order__', axis=1).reset_index(drop=True)


# ----------------------------------------------------------------------------------------------------------
# Pruning the Cache of Deleted Files
# ----------------------------------------------------------------------------------------------------------


def _files_existence_checker(to_check):
    """

    Checks if ``to_check`` exists.
    If not take the following action:

    - If a ``str``, return ``to_check`` if it exists else ``None``.

    - If a ``list`` or ``tuple``, remove items that do not exist.
      If resultant length is zero return ``None``.

    :param to_check: file, or iterable of file, to check the existence of
    :type to_check: ``str``, ``list`` or ``tuple``
    :return: ``to_check``, pruned ``to_check`` (if iterable) or ``None`` (all files removed).
    :rtype: ``str``, ``list``, ``tuple``, ``None`` or ``type(to_check)``
    """
    if isinstance(to_check, str):
        return to_check if os.path.isfile(to_check) else None
    elif isinstance(to_check, (list, tuple)):
        files_present = tuple([i for i in to_check if os.path.isfile(i)])
        return files_present if len(files_present) else None
    else:
        return to_check


def _df_pruner(cache_records_db, columns):
    """

    Prune ``cache_records_db`` by reviewing ``columns``.

    :param cache_records_db: see ``_prune_rows_with_deleted_images()``.
    :type cache_records_db: ``Pandas DataFrame``
    :param columns: see ``_prune_rows_with_deleted_images()``.
    :type columns: ``list``
    :return: a pruned ``cache_records_db``.
    :rtype: ``Pandas DataFrame``
    """
    for c in columns:
        cache_records_db[c] = cache_records_db[c].map(_files_existence_checker)

    # Mark rows to remove
    indices_to_drop = cache_records_db[columns].apply(lambda x: not all(x[i] is None for i in columns), axis=1)

    # Drop and reset the index
    return cache_records_db[indices_to_drop].fillna(np.NaN).reset_index(drop=True)


def _prune_rows_with_deleted_images(cache_records_db, columns, save_path):
    """

    Tool to remove reference to images that have been manually deleted from the cache.
    After this pruning has occurred, ``cache_records_db`` is saved at ``save_path``.

    If a column element is a string, it will be left 'as is' if the file exists,
    otherwise it the entire row will be marked for deletion.

    If a column element is a tuple, image paths in the tuple that do not exist will be removed from the
    tuple. If the resultant tuple is of zero length (i.e., all images have been deleted), the entire row
    will be marked for deletion.

    If, for a given row, all entries for the columns in ``columns`` are ``None`` (i.e., the images
    have been deleted), that row will be removed from ``cache_records_db``.
    Note: if one column is marked for deletion and another is not, the row will be conserved.

    .. note::

        If no images have been deleted, the output of this function will be the same as the input.

    :param cache_records_db: a cache_records_db from the ``OpeniInterface()`` or ``CancerImageInterface()``
    :type cache_records_db: ``Pandas DataFrame``
    :param columns: a ``list`` of columns with paths to cached images. These columns can be columns of
                    strings or columns of tuples.

        .. warning::

            This parameter *must* be a ``list``.

    :type columns: ``list``
    :param save_path: the location to save ``cache_records_db``.
    :type save_path: ``str``
    :return: a pruned ``cache_records_db``
    :rtype: ``Pandas DataFrame``
    """
    pruned_cache_records_db = _df_pruner(cache_records_db, columns)
    pruned_cache_records_db.to_pickle(save_path)
    return pruned_cache_records_db


# ----------------------------------------------------------------------------------------------------------
# Interface Data
# ----------------------------------------------------------------------------------------------------------


_image_instance_image_columns = {
    # Note: the first item should be the default.
    'OpeniInterface': ('cached_images_path',),
    'OpeniImageProcessing': ('cached_images_path',),
    'unify_against_images': ('cached_images_path',),
    'CancerImageInterface': ('cached_dicom_images_path', 'cached_images_path')
}


# ----------------------------------------------------------------------------------------------------------
# Deleting Image Data
# ----------------------------------------------------------------------------------------------------------


def _robust_delete(to_delete):
    """

    Function to delete ``to_delete``.
    If a list (or tuple), all paths therein will be deleted.

    :param to_delete: a file, or multiple files to delete. Note: if ``to_delete`` is not a ``string``,
                     ``list`` or ``tuple``, no action will be taken.
    :type to_delete: ``str``, ``list``  or ``tuple``
    """
    def delete_file(td):
        if os.path.isfile(td):
            os.remove(td)

    if isinstance(to_delete, str):
        delete_file(to_delete)
    elif isinstance(to_delete, (list, tuple)):
        for t in to_delete:
            if isinstance(t, str):
                delete_file(t)


def _double_check_with_user():
    """

    Ask the user to verify they wish to proceed.

    """
    response = input("This action cannot be undone.\n"
                     "Do you wish to continue (y/n)?")
    if cln(response).lower() not in ('yes', 'ye', 'es', 'y'):
        raise ValueError("Action Canceled. 'y' must be entered to proceed.")


def _most_recent_pull_time(instance):
    """
    
    Extract the time of the most recent 
    data pull from ``instance``.
    
    :param instance: see ``image_delete()``
    :type instance: see ``image_delete()``
    :return: see description.
    :rtype: ``bool``
    """
    if instance.__class__.__name__ == 'OpeniImageProcessing':
        records_db = instance.instance.records_db
    else:
        records_db = instance.records_db

    if isinstance(records_db, pd.DataFrame):
        pull_time = records_db['pull_time'].iloc[0]
    else:
        raise TypeError("`only_recent=True` is invalid. The `records_db` "
                        "associated with `instance` is not a DataFrame.")

    return pull_time


def _pretty_print_image_delete(deleted_rows, verbose):
    """

    Pretty print ``deleted_rows``.

    :param deleted_rows: as evolved inside ``image_delete``.
    :type deleted_rows: ``dict``
    :param verbose: see ``image_delete``.
    :type verbose: ``bool``
    """
    if deleted_rows and verbose:
        print("\nIndices of Deleted Rows:\n")
        # Note: `.T` handles when values are of unequal length.
        deleted_rows_df = pd.DataFrame.from_dict(deleted_rows, orient='index').T
        # Put `records_db` before of `cache_records_db` if both are present.
        to_print = deleted_rows_df[sorted(deleted_rows_df.columns, reverse=True)]
        if len(to_print):
            if IN_NOTEBOOK:
                ipython_display(to_print)
            else:
                pandas_pprint(to_print, full_rows=True, suppress_index=True)
        else:
            print("\nNo Rows Deleted.")


def _delete_rule_wrapper_gen(instance, delete_rule, only_recent, image_columns):
    """
    
    Generate ``delete_rule_wrapper()``.
    
    :param instance: see ``image_delete()``
    :type instance: ``OpeniInterface``, ``OpeniImageProcessing`` or ``CancerImageInterface``
    :param delete_rule: see ``image_delete()``
    :type delete_rule: ``str`` or ``func``
    :param only_recent: see ``image_delete()``
    :type only_recent: ``bool``
    :param image_columns: as evolved in ``image_delete()``
    :type image_columns: ``tuple``
    :return: a function which wraps ``delete_rule`` in such a way that a
             boolean is always the output. 
    :rtype: ``func``
    """
    if isinstance(delete_rule, str):
        if cln(delete_rule).lower() == 'all':
            delete_all = True
        else:
            raise ValueError("`delete_rule` must be 'all' or of type 'function'.")
    else:
        delete_all = False

    last_pull_time = _most_recent_pull_time(instance) if only_recent else None

    def delete_rule_wrapper(row, enact):
        """Wrap delete_rule to ensure the output,
         whether or not to delete, is a boolean."""
        if callable(delete_rule):
            if only_recent:
                func_delete = delete_rule(row) and row['pull_time'] == last_pull_time
            else:
                func_delete = delete_rule(row)
        else:
            func_delete = False

        if only_recent:
            blanket_delete = delete_all and row['pull_time'] == last_pull_time
        else:
            blanket_delete = delete_all

        if blanket_delete or (isinstance(func_delete, bool) and func_delete is True):
            if enact:
                for c in image_columns:
                    _robust_delete(row[c])
            return False  # drop row from the dataframe
        else:
            return True   # keep row in the dataframe

    return delete_rule_wrapper


def image_delete(instance, delete_rule, only_recent=False, verbose=True):
    """

    Delete rows, and associated images, from ``records_db`` and ``cache_records_db``
    DataFrames associated with ``instance``.

    .. warning::

        The effects of this function can only be undone by downloading the deleted data again.
        
    .. warning::
    
        The default behavior for this function is to delete *all* rows (and associated image(s))
        that ``delete_rule`` authorizes. To limit deletion to the most recent data pull,
        ``only_recent`` must be set to ``True``.

    :param instance: an instance of ``OpeniInterface``, ``OpeniImageProcessing`` or ``CancerImageInterface``.
    :type instance: ``OpeniInterface``, ``OpeniImageProcessing`` or ``CancerImageInterface``
    :param delete_rule: must be one of: ``'all'`` (delete *all* data) or a ``function`` which (1) accepts a single
                        parameter (argument) and (2) returns ``True`` when the data is to be deleted.
    :type delete_rule: ``str`` or ``func``
    :param only_recent: if ``True``, only apply ``delete_rule`` to data obtained in the most recent pull. Defaults to ``False``.
    :type only_recent: ``bool``
    :param verbose: if ``True``, print additional information.
    :type verbose: ``bool``
    :return: a dictionary of the indices which were dropped. Example: ``{'records_db': [58, 59], 'cache_records_db': [158, 159]}``.
    :rtype: ``dict``

    :Example:

    >>> from biovida.images import image_delete
    >>> from biovida.images import OpeniInterface
    ...
    >>> opi = OpeniInterface()
    >>> opi.search(image_type=['ct', 'mri'], collection='medpix')
    >>> opi.pull()
    ...
    >>> def my_delete_rule(row):
    >>>     if isinstance(row['abstract'], str) and 'Oompa Loompas' in row['abstract']:
    >>>         return True
    ...
    >>> image_delete(opi, delete_rule=my_delete_rule)

    .. note::

        In this example, any rows in the ``records_db`` and ``cache_records_db``
        for which the 'abstract' column contains the string 'Oompa Loompas' will be deleted.
        Any images associated with this row will also be destroyed.

    .. warning::

        If a function is passed to ``delete_rule`` it *must* return a boolean ``True`` to delete a row.
        **All other output will be ignored**.

    """
    # ToDo: refactor. This function is very cumbersome.
    index_dict = dict()
    _double_check_with_user()

    if verbose:
        print("\nDeleting...")

    image_columns = _image_instance_image_columns[instance.__class__.__name__]
    delete_rule_wrapper = _delete_rule_wrapper_gen(instance=instance, delete_rule=delete_rule,
                                                   only_recent=only_recent, image_columns=image_columns)

    def index_dict_update(data_frame_name, stage, data_frame):
        if stage == 'before':
            index_dict[data_frame_name] = {stage: data_frame.index.tolist()}
        elif stage == 'after':
            index_dict[data_frame_name][stage] = data_frame.index.tolist()

    def non_cache_db(data_frame_name, data_frame, enact):
        """Handle `records_db` and `image_dataframe`."""
        index_dict_update(data_frame_name, stage='before', data_frame=data_frame)
        to_conserve = data_frame.apply(lambda r: delete_rule_wrapper(r, enact=enact), axis=1).tolist()
        data_frame = data_frame[to_conserve].reset_index(drop=True)
        index_dict_update(data_frame_name, stage='after', data_frame=data_frame)
        return data_frame, to_conserve

    if instance.__class__.__name__ == 'OpeniImageProcessing':
        instance.image_dataframe, to_conserve = non_cache_db('image_dataframe', instance.image_dataframe, enact=True)
        if instance.db_to_extract == 'records_db':
            instance.instance.records_db = instance.instance.records_db[to_conserve].reset_index(drop=True)
        elif isinstance(instance.instance.records_db, pd.DataFrame):
            index_dict_update('records_db', 'before', data_frame=instance.instance.records_db)
            to_conserve = instance.instance.records_db[image_columns[0]].map(
                lambda x: os.path.isfile(x) if isinstance(x, str) else False)
            instance.instance.records_db = instance.instance.records_db[to_conserve].reset_index(drop=True)
            index_dict_update('records_db', 'after', data_frame=instance.instance.records_db)
        if isinstance(instance.instance.cache_records_db, pd.DataFrame):
            index_dict_update('cache_records_db', 'before', data_frame=instance.instance.cache_records_db)
            instance.instance._load_prune_cache_records_db(load=False)
            index_dict_update('cache_records_db', 'after', data_frame=instance.instance.cache_records_db)
        else:
            raise TypeError("`cache_record_db` is not a DataFrame.")
    else:
        if isinstance(instance.records_db, pd.DataFrame):
            instance.records_db, _ = non_cache_db('records_db', instance.records_db, enact=False)
        if isinstance(instance.cache_records_db, pd.DataFrame):
            index_dict_update('cache_records_db', 'before', data_frame=instance.cache_records_db)
            _ = instance.cache_records_db.apply(lambda r: delete_rule_wrapper(r, enact=True), axis=1)
            instance._load_prune_cache_records_db(load=False)
            instance.cache_records_db = _relationship_mapper(instance.cache_records_db, instance.__class__.__name__)
            instance._save_cache_records_db()
            index_dict_update('cache_records_db', 'after', data_frame=instance.cache_records_db)
        else:
            raise TypeError("`cache_record_db` is not a DataFrame.")

    deleted_rows = {k: sorted(set(v['before']) - set(v['after'])) for k, v in index_dict.items()}
    _pretty_print_image_delete(deleted_rows=deleted_rows, verbose=verbose)

    return deleted_rows


# ----------------------------------------------------------------------------------------------------------
# Divvy Image Data
# ----------------------------------------------------------------------------------------------------------


def _image_divvy_error_checking(divvy_rule, action, train_val_test_dict):
    """

    Check for possible errors for ``image_divvy()``.

    :param divvy_rule: see ``image_divvy()``.
    :type divvy_rule: ``str`` or ``func``
    :param train_val_test_dict: see ``image_divvy()``.
    :type action: ``str``
    :param train_val_test_dict: see ``image_divvy()``.
    :type train_val_test_dict: ``dict``
    """
    if not isinstance(divvy_rule, str) and not callable(divvy_rule):
        raise TypeError("`divvy_rule` must be a string or function.")

    if isinstance(train_val_test_dict, dict):
        if not train_val_test_dict:
            raise KeyError("`train_val_test_dict` is empty")
        if action == 'copy' and 'target_dir' not in train_val_test_dict:
            raise KeyError("`train_val_test_dict` must contain a `target_dir` key "
                           "if `action='copy'`.")
        for k in train_val_test_dict:
            if k not in ('train', 'validation', 'test', 'target_dir', 'delete_source'):
                raise KeyError("Invalid `train_val_test_dict` key: '{0}'.")
        if action == 'ndarray' and 'target_dir' in train_val_test_dict:
            warn("\nThe 'target_dir' entry in `train_val_test_dict` has no\n"
                 "effect when `action='ndarray'`.")


def _robust_copy(to_copy, copy_to, allow_creation, allow_overwrite):
    """

    Function to copy ``to_copy``.
    If a list (or tuple), all paths therein will be copied.

    :param to_copy: a file, or multiple files to delete. Note: if ``to_copy`` is not a ``string``,
                     ``list`` or ``tuple``, no action will be taken.
    :type to_copy: ``str``, ``list``  or ``tuple``
    :param copy_to: the location for the image
    :type copy_to: ``str``
    :param allow_creation: if ``True``, create ``path_`` if it does not exist, else raise.
    :type allow_creation: ``bool``
    :param allow_overwrite: if ``True`` allow existing images to be overwritten. Defaults to ``True``.
    :type allow_overwrite: ``bool``
    """
    _ = directory_existence_handler(path_=copy_to, allow_creation=allow_creation, verbose=True)

    def copy_util(from_path):
        if os.path.isfile(from_path):
            to_path = os.path.join(copy_to, os.path.basename(from_path))
            if not allow_overwrite and os.path.isfile(to_path):
                raise FileExistsError("The following file already exists:\n{0}".format(to_path))
            shutil.copy2(from_path, to_path)
        else:
            warn("No such file:\n'{0}'".format(from_path))

    if isinstance(to_copy, str):
        copy_util(from_path=to_copy)
    elif isinstance(to_copy, (list, tuple)):
        for c in to_copy:
            if isinstance(c, str):
                copy_util(from_path=c)


def _divvy_column_selector(instance, image_column, data_frame):
    """

    Select the column to use when copying images from.

    :param instance:  see ``image_divvy()``
    :type instance: ``OpeniInterface`` or ``CancerImageInterface``
    :param image_column: see ``image_divvy()``
    :type image_column: ``str``
    :param data_frame: as evolved inside  ``image_divvy()``.
    :type data_frame: ``Pandas DataFrame``
    :return: the column in ``data_frame`` to use when copying images to the new location.
    :rtype: ``str``
    """
    if image_column is None:
        return _image_instance_image_columns[type(instance).__name__][0]
    elif not isinstance(image_column, str):
        raise TypeError('`image_column` must be a string or `None`.')
    elif image_column in _image_instance_image_columns[instance.__class__.__name__]:
        if image_column in data_frame.columns:
            return image_column
        else:
            raise KeyError("The '{0}' column is missing from the dataframe.".format(image_column))
    else:
        raise KeyError("'{0}' is not a valid image column.".format(image_column))


def _image_divvy_wrappers_gen(divvy_rule, action, train_val_test_dict, column_to_use, create_dirs, allow_overwrite):
    """

    Wrap the ``divvy_rule`` passed to ``image_divvy()``.

    :param divvy_rule: see ``image_divvy()``.
    :type divvy_rule: ``str`` or ``func``
    :param action: see ``image_divvy()``.
    :type action: ``str``
    :param train_val_test_dict:  see ``image_divvy()``.
    :type train_val_test_dict: ``dict``
    :param column_to_use:  as evolved inside ``image_divvy`` by ``_divvy_column_selector()``
    :type column_to_use: ``str``
    :param create_dirs: see ``image_divvy()``.
    :type create_dirs: ``bool``
    :param allow_overwrite: see ``image_divvy()``.
    :type allow_overwrite: ``bool``
    :return: a function which wraps the function passed to ``divvy_rule()``.
    :rtype: ``func``
    """
    def copy_rule_wrapper(row, copy_to):
        if isinstance(copy_to, (str, tuple, list)):
            all_copy_targets = [copy_to] if isinstance(copy_to, str) else copy_to
            if not len(all_copy_targets):
                return None
            for i in all_copy_targets:
                if not isinstance(i, str):
                    raise TypeError("`divvy_rule` returned an iterable containing "
                                    "a element which is not a string.")
                _robust_copy(to_copy=row[column_to_use], copy_to=i,
                             allow_creation=create_dirs, allow_overwrite=allow_overwrite)
        elif copy_to is not None:
            raise TypeError("String, list or tuple expected. "
                            "`divvy_rule` returned an object of type "
                            "'{0}'.".format(type(copy_to).__name__))

    def divvy_rule_wrapper(row):
        group = divvy_rule(row) if callable(divvy_rule) else divvy_rule
        if not isinstance(group, (str, list, tuple)):
            return None
        if isinstance(group, (list, tuple)) and not len(group):
            return None
        if action == 'copy' and not isinstance(train_val_test_dict, dict):
            copy_rule_wrapper(row, group)
        if isinstance(row[column_to_use], (str, tuple, list)):
            cache_info = [row[column_to_use]] if isinstance(row[column_to_use], str) else list(row[column_to_use])
            if isinstance(group, str):
                return [[os.path.basename(group), cache_info]]
            elif isinstance(group, (list, tuple)):
                return [[os.path.basename(c), cache_info] for c in group]
        else:
            return None

    return divvy_rule_wrapper


def _image_divvy_train_val_test_wrapper(action, verbose, divvy_info, train_val_test_dict):
    """

    Take ``divvy_info`` and pass to ``train_val_test()``

    :param action: see ``image_divvy()``.
    :type action: ``str``
    :param verbose: see ``image_divvy()``.
    :type verbose: ``bool``
    :param divvy_info: as evolved inside ``image_divvy`` (by ``divvy_info_data_frame_to_dict()``).
    :type divvy_info: ``dict``
    :param train_val_test_dict see ``image_divvy()``.
    :type train_val_test_dict:
    :return: see ``train_val_test``.
    :rtype: ``dict``
    """
    random_state = train_val_test_dict.get('random_state', None)
    target = train_val_test_dict.get('target_dir') if action == 'copy' else None
    delete_source = train_val_test_dict.get('delete_source', False) if action == 'copy' else False

    output_dict = train_val_test(data=divvy_info,
                                 train=train_val_test_dict.get('train', None),
                                 validation=train_val_test_dict.get('validation', None),
                                 test=train_val_test_dict.get('test', None),
                                 target_dir=target, action=action, delete_source=delete_source,
                                 random_state=random_state, verbose=verbose)
    return output_dict


def _divvy_info_to_dict(divvy_info):
    """

    Convert the list evolved inside ``divvy_rule_apply()`` in
    ``image_divvy()`` into a dictionary.

    :param divvy_info: a list of the form ``[[string_1, ['a', 'b']], [string_1, ['c', 'd']], [string_2, ['e']], ...]``.
    :type divvy_info: ``list``
    :return: a dictionary form: ``{string_1: ['a', 'b', 'c', 'd'], string_2: ['e'], ...}``.
    :rtype: ``dict``
    """
    d = defaultdict(list)
    for (k, v) in divvy_info:
        d[k] += v
    return dict(d)


def _divvy_openi_image_processing(instance,
                                  divvy_rule,
                                  action,
                                  train_val_test_dict,
                                  create_dirs,
                                  allow_overwrite,
                                  verbose):
    """

    Handle the special case of 'divvying' when an instance of the ``OpeniImageProcessing``
    class is passed to ``image_divvy()``.

    This has to be to be handled separately so that cropping information,
    determined by ``OpeniImageProcessing``'s analysis methods, can be applied.

    :param instance: see ``image_divvy()``.
    :type instance: ``OpeniImageProcessing``
    :param divvy_rule:  see ``image_divvy()``.
    :type divvy_rule: ``str`` or ``func``
    :param action:  see ``image_divvy()``.
    :type action: ``str``
    :param train_val_test_dict: see ``image_divvy()``.
    :type train_val_test_dict: ``dict`` or ``None``
    :param create_dirs:  see ``image_divvy()``.
    :type create_dirs: ``bool``
    :param allow_overwrite: see ``image_divvy()``.
    :type allow_overwrite: ``bool``
    :param verbose: see ``image_divvy()``.
    :type verbose: ``bool``
    :return: yield of ``OpeniImageProcessing.output`` or ``_train_val_test_engine()``.
    :rtype: ``dict``
    """
    if action in ('copy', 'ndarray') and not isinstance(train_val_test_dict, dict):
        output_dict = instance.output(output_rule=divvy_rule, create_dirs=create_dirs,
                                      allow_overwrite=allow_overwrite, action=action)
    elif action in ('copy', 'ndarray') and isinstance(train_val_test_dict, dict):
        if action == 'copy':
            ndarray_with_path = True
            engine_action = 'write_ndarray'
            target_path = train_val_test_dict['target_dir']
        elif action == 'ndarray':
            ndarray_with_path = False
            engine_action = None
            target_path = None

        divvy_info = instance.output(output_rule=divvy_rule,
                                     action='ndarray',  # used so that cropping is applied.
                                     ndarray_with_path=ndarray_with_path)

        output_dict = _train_val_test_engine(action=engine_action,
                                             tvt=_tvt_dict_gen(train_val_test_dict),
                                             group_files_dict=divvy_info,
                                             target_path=target_path,
                                             random_state=train_val_test_dict.get('random_state', None),
                                             allowed_actions=('write_ndarray', None),
                                             verbose=verbose)

    return output_dict


def image_divvy(instance,
                divvy_rule,
                action='ndarray',
                db_to_extract='records_db',
                train_val_test_dict=None,
                create_dirs=True,
                allow_overwrite=True,
                image_column=None,
                stack=True,
                verbose=True):
    """

    Grouping Cached Images.

    .. warning::

        Currently, if an ``OpeniImageProcessing`` instance is first passed to
        ``biovida.unification.unify_against_images`` and then to this function,
        image cropping will not be applied.

    :param instance: the yield of the yield of ``biovida.unification.unify_against_images()`` or an instance of
                     ``OpeniInterface``, ``OpeniImageProcessing`` or ``CancerImageInterface``.
    :type instance: ``OpeniInterface``, ``OpeniImageProcessing``, ``CancerImageInterface`` or ``Pandas DataFrame``
    :param divvy_rule: must be a `function`` which (1) accepts a single parameter (argument) and (2) return
                       system path(s) [see example below].
    :type divvy_rule: ``str`` or ``func``
    :param action: one of: ``'copy'``, ``'ndarray'``.

                    - if ``'copy'``: copy from files from the cache to (i) the location prescribed by ``divvy_rule``,
                      when ``train_val_test_dict=None``, else (ii) the 'target_location' key in ``train_val_test_dict``.

                    - if ``'ndarray'``: return a nested dictionary of ``ndarray`` ('numpy') arrays (default).

    :type action: ``str``
    :param db_to_extract: the database to use. Must be one of:

        * 'records_db': the dataframe resulting from the most recent ``search()`` & ``pull()`` (default).
        * 'cache_records_db': the cache dataframe for ``instance``.
        * 'unify_against_images': the yield of ``biovida.unification.unify_against_images()``.

        .. note::

            If an instance of ``OpeniImageProcessing`` is passed, the dataframe will be extracted automatically.

    :type db_to_extract: ``str``
    :param train_val_test_dict: a dictionary denoting the proportions for any of: ``'train'``, ``'validation'`` and/or ``'test'``.

                        .. note::

                            * If ``action='copy'``, a ``'target_dir'`` key (target directory) *must* also be included.
                            * A ``'random_state'`` key can be passed, with an integer as the value, to seed shuffling.
                            * To delete the source files, a ``'delete_source'`` key may be included (optional).
                              The corresponding value provided *must* be a boolean. If no such key key is provided,
                              ``'delete_source'`` defaults to ``False``.

    :type train_val_test_dict: ``None`` or ``dict``
    :param create_dirs: if ``True``, create directories returned by ``divvy_rule`` if they do not exist. Defaults to ``True``.
    :type create_dirs: ``bool``
    :param allow_overwrite: if ``True`` allow existing images to be overwritten. Defaults to ``True``.
    :type allow_overwrite: ``bool``
    :param image_column: the column to use when copying images. If ``None``, use ``'cached_images_path'``. Defaults to ``None``.
    :type image_column: ``str``
    :param stack: if ``True``, stack 3D volumes and time-series images when ``action='ndarray'``. Defaults to ``True``.
    :type stack: ``bool``
    :param verbose: if ``True`` print additional details. Defaults to ``True``.
    :type verbose: ``bool``
    :return:

        * If ``divvy_rule`` is a string:

          * If ``action='copy'`` and ``train_val_test_dict`` is not dictionary, this function will
            return a dictionary of the form ``{divvy_rule: [cache_file_path, cache_file_path, ...], ...}``.

        * If ``divvy_rule`` is a function:

          * If ``action='copy'`` and ``train_val_test_dict`` is not a dictionary, this function will
            return a dictionary of the form ``{string returned by divvy_rule(): [cache_file_path, cache_file_path, ...], ...}``.

          * If ``action='ndarray'`` and ``train_val_test_dict`` is not a dictionary, this function will
            return a dictionary of the form ``{string returned by divvy_rule(): array([Image Matrix, Image Matrix, ...]), ...}``.

          * If ``train_val_test_dict`` is a dictionary, the output is powered by utilities.train_val_test
            (available :func:`here <biovida.support_tools.utilities.train_val_test>`).

    :rtype: ``dict``

    :Example:

    >>> from biovida.images import image_divvy
    >>> from biovida.images import OpeniInterface

    |
    | **Obtain Images**

    >>> opi = OpeniInterface()
    >>> opi.search(image_type=['mri', 'ct'])
    >>> opi.pull()

    |
    | **Usage 1a**: Copy Images from the Cache to a New Location

    >>> summary_dict = image_divvy(opi, divvy_rule="/your/output/path/here/output", action='copy')

    |
    | **Usage 1b**: Converting to ``ndarrays``

    >>> def my_divvy_rule1(row):
    >>>     if isinstance(row['image_modality_major'], str):
    >>>         if 'mri' == row['image_modality_major']:
    >>>             return 'mri'
    >>>         elif 'ct' == row['image_modality_major']:
    >>>             return 'ct'
    ...
    >>> nd_data = image_divvy(opi, divvy_rule=my_divvy_rule1, action='ndarray')

    The resultant ``ndarrays`` can be extracted as follows:

    >>> ct_images = nd_data['ct']
    >>> mri_images = nd_data['mri']

    |
    | **Usage 2a**: A Rule which Invariably Returns a Single Save Location for a Single Row

    >>> def my_divvy_rule2(row):
    >>>     if isinstance(row['image_modality_major'], str):
    >>>         if 'mri' == row['image_modality_major']:
    >>>             return '/your/path/here/MRI_images'
    >>>         elif 'ct' == row['image_modality_major']:
    >>>             return '/your/path/here/CT_images'
    ...
    >>> summary_dict = image_divvy(opi, divvy_rule=my_divvy_rule2, action='copy')

    |
    | **Usage 2b**: A Rule which can Return Multiple Save Locations for a Single Row

    >>> def my_divvy_rule2(row):
    >>>     locations = list()
    >>>     if isinstance(row['image_modality_major'], str):
    >>>         if 'leg' in row['abstract']:
    >>>             locations.append('/your/path/here/leg_images')
    >>>         if 'pelvis' in row['abstract']:
    >>>             locations.append('/your/path/here/pelvis_images')
    >>>     return locations
    ...
    >>> summary_dict= image_divvy(opi, divvy_rule=my_divvy_rule2, action='copy')

    |
    | **Usage 3**: Divvying into *train/validation/test*

    **i**. Copying to a New Location (reusing ``my_divvy_rule1``)

    >>> train_val_test_dict = {'train': 0.7, 'test': 0.3, 'target_dir': '/your/path/here/output'}
    >>> summary_dict = image_divvy(opi, divvy_rule=my_divvy_rule1, action='copy', train_val_test_dict=train_val_test_dict)

    **ii**. Obtaining ``ndarrays`` (numpy arrays)

    >>> train_val_test_dict = {'train': 0.7, 'validation': 0.2, 'test': 0.1}
    >>> image_dict = image_divvy(opi, divvy_rule=my_divvy_rule1, action='ndarray', train_val_test_dict=train_val_test_dict)

    The resultant ``ndarrays`` can be unpacked as follows:

    >>> train_ct, train_mri = image_dict['train']['ct'], image_dict['train']['mri']
    >>> val_ct, val_mri = image_dict['validation']['ct'], image_dict['validation']['mri']
    >>> test_ct, test_mri = image_dict['test']['ct'], image_dict['test']['mri']

    This function behaves the same if passed an instance of ``OpeniImageProcessing``

    >>> from biovida.images import OpeniImageProcessing
    >>> ip = OpeniImageProcessing(opi)
    >>> ip.auto()
    >>> ip.clean_image_dataframe()

    >>> image_dict = image_divvy(ip, divvy_rule=my_divvy_rule1, action='ndarray', train_val_test_dict=train_val_test_dict)
    ...

    .. note::

        If an instance of ``OpeniImageProcessing`` is passed to ``image_divvy``, the ``image_data_frame_cleaned``
        dataframe will be extracted.

    .. note::

        If a function passed to ``divvy_rule`` returns a system path when a dictionary has been passed
        to ``train_val_test_dict``, only the basename of the system path will be used.

    .. warning::

        While it is possible to pass a function to ``divvy_rule`` which returns multiple categories
        (similar to ``my_divvy_rule2()``) when divvying into *train/validation/test*, doing
        so is not recommended. Overlap between these groups is likely to lead to erroneous
        performance metrics (e.g., accuracy) when assessing fitted models.

    """
    # ToDo: remove need for the first (`OpeniImageProcessing`-related) warning given in the docstring above.
    _image_divvy_error_checking(divvy_rule=divvy_rule, action=action,
                                train_val_test_dict=train_val_test_dict)

    if type(instance).__name__ == 'OpeniImageProcessing':
        return _divvy_openi_image_processing(instance=instance, divvy_rule=divvy_rule,
                                             action=action, train_val_test_dict=train_val_test_dict,
                                             create_dirs=create_dirs, allow_overwrite=allow_overwrite,
                                             verbose=verbose)
    elif db_to_extract == 'unify_against_images':
        data_frame = instance
    else:
        data_frame = getattr(instance, db_to_extract)

    if not isinstance(data_frame, pd.DataFrame):
        raise TypeError("Expected a DataFrame.\n"
                        "Got an object of type: '{0}'."
                        "`db_to_extract` may be invalid".format(type(data_frame).__name__))

    column_to_use = _divvy_column_selector(instance, image_column=image_column, data_frame=data_frame)

    divvy_rule_wrapper = _image_divvy_wrappers_gen(divvy_rule=divvy_rule, action=action,
                                                   train_val_test_dict=train_val_test_dict,
                                                   column_to_use=column_to_use, create_dirs=create_dirs,
                                                   allow_overwrite=allow_overwrite)

    def divvy_rule_apply():
        divvy_info = list()
        for _, row in tqdm(data_frame.iterrows(), total=len(data_frame),
                           desc='Applying Divvy Rule', disable=not verbose):
            target = divvy_rule_wrapper(row)
            if isinstance(target, list):
                for t in target:
                    divvy_info.append(t)
        return _divvy_info_to_dict(divvy_info) if len(divvy_info) else None

    divvy_info = divvy_rule_apply()

    if isinstance(divvy_info, dict) and isinstance(train_val_test_dict, dict):
        return _image_divvy_train_val_test_wrapper(action=action, verbose=verbose, divvy_info=divvy_info,
                                                   train_val_test_dict=train_val_test_dict)
    elif isinstance(divvy_info, dict) and action == 'ndarray':
        return _file_paths_dict_to_ndarrays(divvy_info, dimensions=1, stack=stack, verbose=verbose)
    elif isinstance(divvy_info, dict) and action == 'copy':
        return divvy_info
